'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var zora = require('zora');
var util = require('util');

const REPORT_HOOKS = Symbol('@zorax/plug:REPORT_HOOKS');

const noop$1 = () => {};

const getHarnessMutators = pg => [pg.test, pg.harness, pg.init];

const getHarnessDecorators = pg => [
  pg.decorate,
  pg.decorateHarness,
  pg.decorateInit,
];

const getProxyMutators = pg => [pg.test, pg.harness];

const getProxyDecorators = pg => [pg.decorate, pg.decorateHarness];

const getTestMutators = pg => pg.test;

const getTestDecorators = pg => pg.decorate;

const getHarnessHooks = plugins =>
  [
    ...plugins.flatMap(getHarnessMutators),
    ...plugins.flatMap(getHarnessDecorators),
  ].filter(Boolean);

const getTestHooks = plugins =>
  [...plugins.map(getTestMutators), ...plugins.map(getTestDecorators)].filter(
    Boolean
  );

const getProxyHooks = plugins =>
  [
    ...plugins.flatMap(getProxyMutators),
    ...plugins.flatMap(getProxyDecorators),
  ].filter(Boolean);

const wrapTestMethod = (harness, proxy, withHooks, target) => {
  const { test } = target;

  target.test = (desc, run, ...rest) => {
    // just a friendly sanity check: end of pipeline respects expected signature
    if (
      // undefined title is actually more or less allowed
      (typeof desc !== 'undefined' && typeof desc !== 'string') ||
      typeof run !== 'function'
    ) {
      throw new Error('plugins pipeline in disarray')
    }

    const zora_spec_fn = (t, ...rest) => run(withHooks(t), ...rest);

    return test(desc, zora_spec_fn, ...rest)
  };
};

// NOTE prop: {} makes it undefined
const proxySpecifiers = {
  report: {},
  pass: { writable: true }, // zora.pass wants to rewrite this
  error: {},
  length: {},
  count: {},
  successCount: {},
  skipCount: {},
  failureCount: {},
};

const augmentProxy = (harness, target, plugins, getHooks) => {
  target.plugins = plugins;

  target.plug = (...newPlugins) => plug$1(harness, plugins, target, newPlugins);

  target.run = fn => {
    fn(target);
    return target
  };

  const testHooks = getTestHooks(plugins);

  const withTestHooks = t => {
    wrapTestMethod(harness, target, withTestHooks, t);
    testHooks.forEach(hook => hook(t, target, harness));
    return t
  };

  wrapTestMethod(harness, target, withTestHooks, target);

  // NOTE apply hooks _after_ replacing test!
  const hooks = getHooks(plugins);
  hooks.forEach(hook => hook(target, harness));
};

const createProxy = t => Object.create(t, proxySpecifiers);

const plug$1 = (harness, parentPlugins, target, newPluginsInput) => {
  const newPlugins = newPluginsInput.flat().filter(Boolean);

  newPlugins.forEach(pg => {
    if (pg.init || pg.decorateInit) {
      throw new Error(
        'Init plugin can only be added at harness creation (not plugged): ' +
          pg.name
      )
    }

    if (pg.report) {
      harness[REPORT_HOOKS].push(pg.report);
    }
  });

  const proxy = createProxy(target);

  const plugins = [...parentPlugins, ...newPlugins];

  augmentProxy(harness, proxy, plugins, getProxyHooks);

  return proxy
};

const parseCreateHarnessFactoryArgs = args => {
  if (args.length === 0) {
    return args
  }

  let createHarness;
  let options;
  let plugins;

  const [a, b, c] = args;

  if (typeof a === 'function') {
    createHarness = a;
    if (Array.isArray(b)) {
      plugins = b;
    } else {
      options = b;
      plugins = c;
    }
  } else {
    if (Array.isArray(a)) {
      plugins = a;
    } else {
      options = a;
      plugins = b;
    }
  }

  return [createHarness, options, plugins]
};

// createHarnessFactory(options)
// createHarnessFactory(plugins)
// createHarnessFactory(options, plugins)
// createHarnessFactory(createHarness, options)
// createHarnessFactory(createHarness, plugins)
// createHarnessFactory(createHarness, options, plugins)
const createHarnessFactory = (...args) => {
  const [
    createHarness = zora.createHarness,
    factoryOptions = {},
    factoryPlugins = [],
  ] = parseCreateHarnessFactoryArgs(args);

  const createZoraxHarness = (
    harnessOptions = {},
    harnessPlugins = [],
    ...args
  ) => {
    if (Array.isArray(harnessOptions)) {
      harnessPlugins = harnessOptions;
      harnessOptions = {};
    }

    const options = { ...factoryOptions, ...harnessOptions };

    const harness = createHarness(options, ...args);

    harness.options = options;

    const plugins = [...factoryPlugins, ...harnessPlugins].filter(Boolean);

    harness[REPORT_HOOKS] = plugins.filter(pg => pg.report).map(pg => pg.report);

    const { report } = harness;
    harness.report = async (...args) => {
      const after = [];
      for (const reportHook of harness[REPORT_HOOKS]) {
        const afterHook = await reportHook(harness, args);
        if (afterHook) after.push(afterHook);
      }

      await report.apply(harness, args);

      for (const afterHook of after) {
        await afterHook();
      }
    };

    augmentProxy(harness, harness, plugins, getHarnessHooks);

    return harness
  };

  return createZoraxHarness
};

const createHarness = createHarnessFactory(zora.createHarness);

createHarness.plug = noop$1;

/**
 * Zorax alias
 */

const applyAlias = (from, t) => to => {
  const path = to.split('.');
  const key = path.pop();
  const target = path.reduce((cursor, step) => cursor[step], t);
  target[key] = t[from];
};

const forEach = (run, source) => {
  Array.isArray(source) ? source.forEach(run) : run(source);
};

const applyAliases = (aliases, t) => {
  Object.entries(aliases).forEach(([from, to]) => {
    forEach(applyAlias(from, t), to);
  });
};

const aliaser = aliases => aliases && (t => applyAliases(aliases, t));

var withAliases = (test, harness) => ({
  name: 'zorax.alias',
  decorate: aliaser(test),
  decorateHarness: aliaser(harness),
});

const logError = err => {
  // eslint-disable-next-line no-console
  console.error((err && err.stack) || err);
};

var withAuto = ({
  auto: defaultAuto = true,
  onReportError = logError,
} = {}) => ({
  name: 'zorax.auto',

  init: (t, { options: { auto = defaultAuto } }) => {
    const report = t.report;

    let autoStart = auto;

    t.auto = (enable = autoStart) => {
      autoStart = enable;
      return autoStart
    };

    t.report = (...args) => {
      autoStart = false;
      return report.apply(t, args)
    };

    const maybeStart = () => {
      if (autoStart) {
        t.report().catch(onReportError);
      }
    };

    setTimeout(maybeStart);
  },
});

const specFnRegexp = /zora_spec_fn/;

const zoraInternal = /zora\/dist\/bundle/;

const filterStackLine = l =>
  (l && !zoraInternal.test(l) && !l.startsWith('Error')) || specFnRegexp.test(l);

const getAssertionLocation = err => {
  const stack = (err.stack || '')
    .split('\n')
    .map(l => l.trim())
    .filter(filterStackLine);
  const userLandIndex = stack.findIndex(l => specFnRegexp.test(l));
  const stackline =
    userLandIndex >= 1 ? stack[userLandIndex - 1] : stack[0] || 'N/A';
  return stackline.replace(/^at|^@/, '')
};

const wrapSpec = spec =>
  async function zora_spec_fn(t, ...specRest) {
    try {
      return await spec(t, ...specRest)
    } catch (error) {
      if (error.fatal) {
        throw error
      }
      const at = getAssertionLocation(error);
      t.collect({
        pass: false,
        actual: error,
        expected: 'no thrown error',
        description: 'test should not throw',
        operator: 'doesNotThrow',
        get at() {
          return at
        },
        set at(value) {},
      });
    }
  };

var withCatch = () => ({
  name: 'zorax.catch',
  test: t => {
    const { test } = t;
    t.test = async (title, spec, ..._) => test(title, wrapSpec(spec), ..._);
  },
});

const isFunction = x => typeof x === 'function';

const hasName = name => ({ name: x }) => x === name;

const noop = () => {};

const Deferred = () => {
  let resolve, reject;
  const promise = new Promise((_resolve, _reject) => {
    resolve = _resolve;
    reject = _reject;
  });
  return { promise, resolve, reject }
};

// fatal errors cause a bailout instead of reporting failure with zorax.catch
class FatalError extends Error {
  constructor(...args) {
    super(...args);
    this.fatal = true;
  }
}

const ZORAX_DEFER = 'zorax.defer';
const ZORAX_DEFER_PRINT = 'zorax.defer.print';
const ZORAX_GROUP = 'zorax.defer.group';
const ZORAX_ONLY = 'zorax.defer.only';
const ZORAX_MACRO = 'zorax.macro';

/**
 * zorax/defer
 */

var withDefer = () => ({
  name: ZORAX_DEFER,

  init: (h, { plugins }) => {
    const mainTest = h.test;
    const mainSkip = h.skip;

    let tests = [];

    let running = false;
    let reporting = false;

    const pushTest = spec => {
      tests.push(spec);
    };

    const runTestInCtx = ({ test, skip }) => spec => {
      const { args, resolve, reject, skip: skipped } = spec;
      const run = skipped ? skip : test;
      run(...args)
        .then(resolve)
        .catch(err => {
          reject(err);
          throw err
        });
    };

    const deferPlugins = plugins.map(pg => pg.defer).filter(Boolean);

    const addHooks = deferPlugins.map(defer => defer.add).filter(Boolean);

    const runHooks = deferPlugins.map(defer => defer.run).filter(Boolean);

    const _addTest = addHooks.reduce((push, hook) => hook(push, h), pushTest);

    const runTestIn = runHooks.reduce((run, hook) => hook(run, h), runTestInCtx);

    const addTest = (args, extra) => {
      const deferred = Deferred();
      deferred.args = args;
      _addTest(deferred);
      Object.assign(deferred, extra);
      return deferred.promise
    };

    // h.defer.runner(t)
    //   => (spec => { /* run */ })
    //
    // Allows a plugin to run specs through the entire run pipeline, not just
    // what's before them.
    //
    // This is necessary for a plugin that "unfolds" new test items (previous
    // plugins won't have seen them, since they enter "from the middle").
    //
    // defer         A             B
    //              if (...) <--- filter <----------------
    //                |                                   |
    //                |                  defer.runner(t)( o )
    //                |                                   |
    //               then --[o,o,o]-- unfold -- o,o, ----
    //                |
    //  run(o) <--- else    h.defer.runner(t)
    //
    h.defer = {
      runner: runTestIn,
    };

    const flush = () => {
      running = true;
      const runSpec = runTestIn({ test: mainTest, skip: mainSkip });
      const currentTests = tests;
      tests = [];
      for (const spec of currentTests) {
        runSpec(spec);
      }
      running = false;
    };

    const deferTest = (args, tester = mainTest, extra) => {
      // guard: is passthrough during reporting
      if (running) {
        return tester(...args)
      }
      // guard: can't add top level test during reporting
      if (reporting) {
        throw new Error('Cannot add test after reporting has started')
      }
      addTest(args, extra);
    };

    h.test = (...args) => deferTest(args);

    h.skip = (...args) => deferTest(args, mainSkip, { skip: true });

    const _report = h.report;
    h.report = async (...args) => {
      reporting = true;
      flush();
      const result = await _report.apply(h, args);
      reporting = false;
      return result
    };
  },
});

const pass = value => fn => fn(value);

const matcher = pattern => {
  const patternLc = pattern.toLowerCase();
  return x => x && x.toLowerCase().includes(patternLc)
};

const notMatcher = pattern => {
  const patternLc = pattern.toLowerCase();
  return x => !(x && x.includes(patternLc))
};

const createFilter = filter => {
  if (filter == null) return null
  if (Array.isArray(filter)) {
    const include = [];
    const adds = [];
    const subs = [];
    filter.flat().forEach(pattern => {
      const mod = pattern.slice(0, 1);
      if (mod === '+') {
        adds.push(matcher(pattern.slice(1)));
      } else if (mod === '-') {
        subs.push(matcher(pattern.slice(1)));
      } else {
        include.push(matcher(pattern));
      }
    });
    return x => {
      const passes = pass(x);
      if (subs.length && subs.some(passes)) {
        return false
      }
      if (adds.length && !adds.every(passes)) {
        return false
      }
      if (include.length && !include.some(passes)) {
        return false
      }
      return true
    }
  }
  if (filter.slice(0, 1) === '!') {
    const pattern = filter.slice(1);
    return notMatcher(pattern)
  }
  const isMatch = matcher(filter);
  // for debug
  isMatch.input = filter;
  return isMatch
};

var withFilter = () => {
  let isMatch = null;

  return {
    name: 'zorax.filter',

    description: 'filter test by title',

    harness(h, { options: { skip: useSkip = false } }) {
      h.filter = filter => {
        isMatch = createFilter(filter);
      };

      const { test, skip } = h;
      h.test = async (desc, ...args) => {
        if (isMatch != null && !isMatch(desc)) {
          if (useSkip) {
            return skip(desc, ...args)
          }
          return
        }
        return test(desc, ...args)
      };
    },
  }
};

const TITLE_ONLY = Symbol('TITLE_ONLY');

const enforceNoAsync = result => {
  if (result && isFunction(result.then) && isFunction(result.catch)) {
    throw new Error('group / describe / it handlers cannot be async')
  }
};

const isEmpty = tests =>
  tests.every(item => {
    if (Array.isArray(item)) {
      return isEmpty(item[1])
    } else {
      return false
    }
  });

const validateDependencies = ({ plugins }) => {
  // require defer
  const deferIndex = plugins.findIndex(hasName(ZORAX_DEFER));
  if (deferIndex === -1) {
    throw new Error(`${ZORAX_GROUP} requires ${ZORAX_DEFER}`)
  }
  // must be after defer
  const index = plugins.findIndex(hasName(ZORAX_GROUP));
  if (index < deferIndex) {
    throw new Error(`${ZORAX_GROUP} must be after ${ZORAX_DEFER}`)
  }
};

var withGroup = () => ({
  name: ZORAX_GROUP,

  harness: validateDependencies,

  defer: {
    add: (_add, harness) => {
      const stack = [];
      let cursor = null;
      let skipping = 0;

      const add = spec => {
        spec.skip = skipping > 0;
        if (cursor) {
          cursor.push(spec);
        } else {
          _add(spec);
        }
      };

      const updateCursor = () => {
        cursor = stack[stack.length - 1];
      };

      const push = (title, resolve) => {
        const tests = [];
        tests.resolve = resolve;
        add([title, tests]);
        tests.title = title;
        stack.push(tests);
        updateCursor();
      };

      const pop = () => {
        const item = stack.pop();
        if (item && item.resolve) {
          item.resolve();
        }
        updateCursor();
      };

      const reset = () => {
        while (stack.length > 0) {
          pop();
        }
      };

      let deep = 0;

      // NOTE callback is used internally for skipping
      const group = (callback, title, handler) => {
        // enable: group(() => { ... })
        if (typeof title === 'function') {
          enforceNoAsync(title());
          if (callback) callback();
          return
        }
        // enable: group.only() -- top level
        if (title === TITLE_ONLY) {
          title = cursor.title + ' <<< ONLY';
        }
        if (handler) {
          deep++;
          push(title, callback);
          enforceNoAsync(handler());
          pop();
          deep--;
        } else {
          if (deep > 0) {
            throw new FatalError(
              "group('...') with no handler is only allowed at top level"
            )
          }
          reset();
          push(title, callback);
        }
      };

      const skip = (callback, ...args) => {
        // enable: group.skip(() => { ... })
        if (typeof args[0] !== 'string') {
          return
        }
        skipping++;
        return group(() => {
          if (callback) callback();
          skipping--;
        }, ...args)
      };

      harness.group = (title, handler) => group(null, title, handler);

      harness.group.skip = (title, handler) => skip(null, title, handler);

      // used by defer.only
      harness.group.callback = group;
      harness.group.skip.callback = skip;

      return add
    },

    run: (prev, h) => {
      const makeRunWithGroup = t => {
        const runSpec = prev(t);

        return spec => {
          if (!Array.isArray(spec)) {
            return runSpec(spec)
          }

          const [title, subTests] = spec;

          // guard: no sub tests -> do not create an empty t.test
          if (isEmpty(subTests)) return false

          const handler = tt => {
            const runner = makeRunWithGroup;
            tt.offset = (t.offset || 0) + 1;
            subTests.forEach(runner(tt));
          };

          return runSpec({
            args: [title, handler],
            skip: spec.skip,
            group: true,
          })
        }
      };
      return makeRunWithGroup
    },
  },
});

/**
 * zorax/macro
 */

const wrap = tester => {
  return (...args) => {
    const providedTitle =
      typeof args[0] === 'string' || !args[0] ? args.shift() : undefined;

    const [run, ...data] = args;

    // case: t.skip('no spec function!')
    if (args.length < 1) {
      return tester(providedTitle)
    }

    const title = run.title
      ? typeof run.title === 'string'
        ? run.title
        : run.title(providedTitle, ...data)
      : providedTitle;

    const zora_spec_fn = t => run(t, ...data);

    return tester(title, zora_spec_fn)
  }
};

var withMacro = () => ({
  name: ZORAX_MACRO,

  test: t => {
    const { test, skip, only } = t;

    t.test = wrap(test);

    t.skip = wrap(skip);

    t.only = wrap(only);
  },
});

const removeNonOnlySubs = (() => {
  const hasOnly = sub => sub.only;

  const removeNonOnlySubsFilter = sub =>
    Array.isArray(sub) ? removeNonOnlySubs(sub) : hasOnly(sub);

  const ONLY_SUBS_FILTERED = Symbol('ONLY_SUBS_FILTERED');

  const removeNonOnlySubs = spec => {
    if (spec[ONLY_SUBS_FILTERED]) {
      return spec[1].length > 0
    }

    const filteredSubs = spec[1].filter(removeNonOnlySubsFilter);

    spec[ONLY_SUBS_FILTERED] = true;
    Object.freeze(filteredSubs);

    spec[1] = filteredSubs;

    return filteredSubs.length > 0
  };

  return removeNonOnlySubs
})();

const wrapRun = (run, args, callback) => {
  if (run.callback) {
    return run.callback(callback, ...args)
  } else {
    const result = run(...args);
    callback();
    return result
  }
};

const validateDeps$1 = ({ plugins }) => {
  // require zorax.defer
  const deferIndex = plugins.findIndex(hasName(ZORAX_DEFER));
  if (deferIndex === -1) {
    throw new Error(`${ZORAX_ONLY} requires ${ZORAX_DEFER}`)
  }
  // must be after zorax.defer
  const index = plugins.findIndex(hasName(ZORAX_ONLY));
  if (index < deferIndex) {
    throw new Error(`${ZORAX_ONLY} must be after ${ZORAX_DEFER}`)
  }
  // must be after zorax.group, if zorax.group is present
  const groupIndex = plugins.findIndex(hasName(ZORAX_GROUP));
  if (index < groupIndex) {
    throw new Error(`${ZORAX_ONLY} must be after ${ZORAX_GROUP}`)
  }
};

var withOnly = () => {
  let hasOnly = false;
  let running = false;

  let only = 0;

  const makeOnly = (run, h) => (title, ...args) => {
    if (running) {
      throw new FatalError('defer.only: only can only be called synchronously')
    }
    // NOTE late resolution of options.only to enable on the fly change
    if (!h.options.only) {
      throw new FatalError(
        'defer.only: only is only authorized when passing the only flag'
      )
    }
    only++;
    return wrapRun(run, [title || TITLE_ONLY, ...args], () => {
      only--;
    })
  };

  return {
    name: ZORAX_ONLY,

    defer: {
      add: _add => spec => {
        _add(spec);

        spec.only = !spec.skip && only > 0;

        if (spec.only) hasOnly = true;
      },

      run: prev => t => {
        const runTest = prev(t);

        if (!hasOnly) {
          return runTest
        }

        return spec => {
          if (spec.only === false) {
            spec.skip = true;
            return false
          }

          // prevent groups that are emptied by only to appear non empty
          if (Array.isArray(spec)) {
            // NOTE mutating spec because not sure what prop (e.g. .skip,
            // .resolve) it might have
            const empty = !removeNonOnlySubs(spec);
            if (empty) {
              return false
            }
          }

          return runTest(spec)
        }
      },
    },

    test(t) {
      t.only = () => {
        throw new FatalError(
          'defer.only: only is only available for root tests'
        )
      };
    },

    harness(t, h) {
      validateDeps$1(t);
      const { test } = t;
      t.only = makeOnly(test, h);
    },

    decorateHarness(t, h) {
      const { group } = t;
      if (group) {
        t.group.only = makeOnly(group, h);
      }
    },

    init(h) {
      const { report } = h;
      h.report = async (...args) => {
        running = true;
        const result = await report.call(h, ...args);
        running = false;
        return result
      };
    },
  }
};

// https://stackoverflow.com/a/41407246/1387519
const bold = '\x1b[1m';
const reset = '\x1b[0m';
const cyan = '\x1b[36m';
const magenta = '\x1b[35m';
// const blue = '\x1b[34m'

const write = console.log.bind(console);

const printer = handler => ({
  compact: initialCompact = false,
  indent: customIndent,
} = {}) => {
  let maxEmpties;
  let indentString = customIndent;
  let compact;

  const setCompact = _compact => {
    compact = _compact;
    maxEmpties = compact ? 0 : 1;
    if (!customIndent) {
      indentString = compact ? '  ' : '    ';
    }
  };

  setCompact(initialCompact);

  let lastEmpties = 0;

  const _log = (_indent, color, prefix, ...args) => {
    const value = args.join(' ');

    const empty = value === '';
    if (empty) {
      if (lastEmpties < maxEmpties) {
        lastEmpties++;
        return write()
      }
      return
    }
    lastEmpties = 0;

    const indent = indentString.repeat(_indent);
    const formatted = indent + value;
    return color
      ? write(`%s${color}%s\x1b[0m`, indent + prefix + ' ', value)
      : write(formatted)
  };

  let testCount = 0;
  let groupCount = 0;

  const logTest = (offset, title, isGroup) => {
    if (isGroup) {
      groupCount++;
      _log();
    } else {
      testCount++;
    }
    _log(offset, isGroup ? bold + magenta : cyan, isGroup ? '▹' : '•', title);
    if (isGroup) {
      _log();
    }
  };

  let enabled = false;

  const isEnabled = () => enabled;

  const { init, ...props } = handler({ isEnabled, logTest });

  return {
    init(h) {
      const {
        report,
        options: { printCompact = compact },
      } = h;

      setCompact(printCompact);

      h.print = arg => {
        if (arg) {
          enabled = true;
          if (arg.hasOwnProperty('compact')) {
            setCompact(arg.compact);
          }
        } else {
          enabled = false;
        }
      };

      const plural = x => (x === 1 ? '' : 's');

      h.report = async (...args) => {
        if (enabled) {
          const space = compact ? '' : '\n';
          write(
            space +
              `${bold + cyan + testCount} test${plural(testCount)} ${reset}` +
              `${magenta}(${groupCount} group${plural(groupCount)})${reset}` +
              space
          );
          testCount = groupCount = 0;
          return { skip: true }
        }
        return report(...args)
      };

      if (init) {
        init(h);
      }
    },

    ...props,
  }
};

const validateDeps = ({ plugins }) => {
  // require zorax.defer
  const deferIndex = plugins.findIndex(hasName(ZORAX_DEFER));
  if (deferIndex === -1) {
    throw new FatalError(`${ZORAX_DEFER_PRINT} requires ${ZORAX_DEFER}`)
  }
  // must be before zorax.defer
  const index = plugins.findIndex(hasName(ZORAX_DEFER_PRINT));
  if (index > deferIndex) {
    throw new FatalError(`${ZORAX_DEFER_PRINT} must be before ${ZORAX_DEFER}`)
  }
};

var withPrint = printer(({ isEnabled, logTest }) => ({
  name: 'zorax.defer.print',

  description: "print tests' title instead of running them",

  init: validateDeps,

  defer: {
    run: prev => t => {
      const runSpec = prev(t);
      if (!isEnabled()) {
        return runSpec
      }
      return spec => {
        const offset = t.offset;
        const title = spec.args && spec.args[0];
        const isGroup = spec.group === true;

        logTest(offset, title, isGroup);

        return isGroup ? runSpec(spec) : runSpec({ ...spec, skip: true })
      }
    },
  },
}));

/**
 * zorax/pass
 */

var withPass = () => ({
  name: 'zorax.pass',
  test: t => {
    t.pass = (msg = 'pass called') => t.ok(true, msg);
  },
});

var withReporter = (
  defaultNoIndentReporter,
  defaultIndentReporter = defaultNoIndentReporter
) => ({
  name: 'zorax.reporter',

  init: (t, { options: { reporter: optionsReporter, indent = false } }) => {
    const report = t.report;

    const defaultReporter = optionsReporter
      ? optionsReporter
      : indent
      ? defaultIndentReporter || zora.mochaTapLike
      : defaultNoIndentReporter || zora.tapeTapLike;

    t.report = (reporter = defaultReporter, ...args) =>
      report.call(t, reporter, ...args);
  },
});

const times = n => `${n} ${n === 1 ? 'time' : 'times'}`;

const passed = { pass: true };

const failed = {
  pass: false,
  with: () => failed,
  returned: () => failed,
};

const fastFormat = value =>
  util.inspect(value, {
    depth: 1,
    maxArrayLength: 3,
    breakLength: Infinity,
  });

const formatArgs = value => `(${value.map(fastFormat).join(', ')})`;

const spacedIf = name => name && `${name.trim()} `;

const ordinal = i => {
  const x = String(i + 1);
  switch (x) {
    case '1':
      return '1st'
    case '2':
      return '2nd'
    case '3':
      return '3rd'
    default:
      return `${x}th`
  }
};

const createSpyFactory = t => (spyName, fn = noop) => {
  if (typeof spyName !== 'string') {
    fn = spyName || noop;
    spyName = fn.name || '';
  }
  const calls = [];
  const _args = [];
  const returns = [];
  const nextCallListeners = [];

  const spacedSpyName = spacedIf(spyName);

  const spied = function zora_spec_fn(...args) {
    const result = fn.apply(this, args);
    const call = [args, result];
    spied[calls.length] = call;
    calls.push(call);
    _args.push(args);
    returns.push(result);
    flushNextCall();
    return result
  };

  // useful for debug
  spied.fn = fn;

  Object.assign(spied, { calls, args: _args, returns });

  const beenCalledWith = (args, api = passed) => {
    const { pass } = t.eq(
      _args,
      args,
      `${spacedSpyName}should have been called with ${args
        .map(formatArgs)
        .join(' then ')}`
    );
    return pass ? api : failed
  };

  const hasReturned = (results, api = passed) => {
    const { pass } = t.eq(
      returns,
      results,
      `${spacedSpyName}should have returned ${results
        .map(fastFormat)
        .join(' then ')}`
    );
    return pass ? api : failed
  };

  spied.hasBeenCalled = (...args) => {
    if (Array.isArray(args[0])) {
      t.eq(
        calls,
        args,
        `${spacedSpyName}should have been called with ${formatArgs(args)}`
      );
    } else {
      const [
        n,
        msg = `${spacedSpyName}should have been called ${times(n)}`,
      ] = args;
      t.eq(calls.length, n, msg);
    }
    const api = {
      with: (...args) => beenCalledWith(args, api),
      returned: (...results) => hasReturned(results, api),
    };
    return api
  };

  const calledWith = (i, args, api = passed, name = spyName) => {
    const { pass } = t.eq(
      _args[i],
      args,
      `${spacedIf(name)}should be called ${ordinal(i)} with ${formatArgs(args)}`
    );
    return pass ? api : failed
  };

  const returned = (i, result, api = passed, name = spyName) => {
    const { pass } = t.eq(
      returns[i],
      result,
      `${spacedIf(name)}${ordinal(i)} call should return ${fastFormat(result)}`
    );
    return pass ? api : failed
  };

  let cur = 0;

  spied.wasCalled = (...params) => {
    const [index, args, result] = params;
    const hasResult = params.length > 2;
    const n = calls.length;
    let i = index;
    const hasName = typeof i === 'string';
    const name = hasName ? i + ' ' : spyName;
    let r = { pass: true };
    if (i == null || hasName) {
      i = cur++;
      const msg =
        n > i
          ? `${spacedIf(name)}was called ${times(cur)} before expected`
          : `${spacedIf(name)}should have been called ${times(i + 1)}`;
      r = t.eq(n, i + 1, msg);
    } else {
      r = t.ok(n > i, `should be called at least ${times(i + 1)}`);
    }
    if (r.pass && args) {
      r = calledWith(i, args);
    }
    if (r.pass && hasResult) {
      r = returned(i, result);
    }
    if (!r.pass) {
      return failed
    }
    const api = {
      with: (...args) => calledWith(i, args, api, name),
      returned: result => returned(i, result, api, name),
    };
    return api
  };

  spied.wasNotCalled = msg => {
    t.ok(
      calls.length === cur,
      msg || `${spacedSpyName}should not have been called`
    );
  };

  spied.wasCalledWith = (...args) => spied.wasCalled().with(...args);

  spied.returned = result => spied.wasCalled().returned(result);

  // === modifiers ===

  // --- .just. ---

  spied.just = {
    wasCalled: (...args) => {
      if (args.length > 0) {
        throw new Error('just.wasCalled is only allowed with zero arguments')
      }
      cur = calls.length - 1;
      return spied.wasCalled()
    },
    wasCalledWith: (...args) => spied.just.wasCalled().with(...args),
    returned: (...args) => spied.just.wasCalled().returned(...args),
  };

  // --- quantifier: spy.o.o.x ---

  const X = i => ({
    wasCalledWith: (...args) => spied.wasCalled(i).with(...args),
    returned: result => spied.wasCalled(i).returned(result),
  });

  spied.x = X(0);

  Object.defineProperty(spied, 'o', {
    get() {
      let i = 1;
      const o = {
        get o() {
          i++;
          return o
        },
        get x() {
          return X(i)
        },
      };
      return o
    },
  });

  // --- .first ---

  spied.first = X(0);

  // --- .last ---

  spied.last = {
    wasCalledWith: (...args) => spied.wasCalled(calls.length - 1).with(...args),
    returned: (...args) => spied.wasCalled(calls.length - 1).returned(...args),
  };

  // --- .nextCall() ---

  spied.nextCall = () =>
    new Promise(resolve => {
      nextCallListeners.push(resolve);
    });

  spied.nextCallAfter = async op => {
    const called = spied.nextCall();
    await op();
    await called;
  };

  const flushNextCall = () => {
    const flushed = nextCallListeners.splice(0);
    setTimeout(() => {
      for (const fn of flushed) {
        fn();
      }
    });
  };

  return spied
};

const spiesTrap = (t, o) => prop => {
  o[prop] = (...args) => (o[prop] = t.spy(...args));
};

const createSpiesFactory = t => (...names) => {
  const spies = {};
  const trap = spiesTrap(t, spies);
  if (names.length > 0) {
    names.forEach(trap);
    return spies
  }
  return new Proxy(spies, {
    get(target, prop) {
      if (!target.hasOwnProperty(prop)) {
        trap(prop);
      }
      return target[prop]
    },
  })
};

var withSpy = () => ({
  name: 'zorax.spy',
  test(t) {
    t.spy = createSpyFactory(t);
    t.spies = createSpiesFactory(t);
  },
});

/**
 * zorax.todo
 *
 * TODO
 */

var withTodo = () => ({
  name: 'zorax.todo',
  decorate: t => {
    t.test.todo = title => t.skip(title);
  },
});

var createDefaultHarness = createHarnessFactory([
  withReporter(),

  withAuto({ auto: true }),

  withCatch(),

  withPrint(),

  withDefer(),
  withGroup(),
  withOnly(),

  withPass(),
  withTodo(),
  withSpy(),

  withFilter(),

  withMacro(),

  withAliases({
    group: 'describe',
    only: 'test.only',
    skip: 'test.skip',
  }),
]);

const harness = createDefaultHarness();

// export const createHarness = (...args) => {
//   harness.auto(false)
//   return createDefaultHarness(...args)
// }
//

const { test, group, describe, plug } = harness;

exports.describe = describe;
exports.group = group;
exports.harness = harness;
exports.plug = plug;
exports.test = test;
//# sourceMappingURL=zorax.js.map
